/******** <lalVerbatim file="BlockApplySearchCV"> ********
Author: Tibbits, M. M.
$Id: BlockApplySearch.c,v 1.30 2003/07/06 20:06:08 jolien Exp $
********* </lalVerbatim> ********/

/******************************* <lalLaTeX file="BlockApplySearchC">
\subsection{Module \texttt{BlockApplySearch.c}}
\label{s:BlockApplySearch.c}

Routine that uses parallelism of processes to implement the rho2 backround
search.  The rho2 search is based on, a yet to be published, block
normailization paper.  Our search returns a result or "expectation" value
which corresponds to the probability of there being multiple gaussian
waveforms within the measured data.  The search also returns the index
and magnitude of the maximum value in the input vector.  Via a filterparam
you can also enable the Rho3 search which chops the input at the maximum
value and reruns the Rho2 code on both segements.  Thereby, we can search
for both the beginning and the end of a gaussian waveform present.

\subsubsection*{Prototypes}
\input{BlockApplySearchCP}
\index{\texttt{LALApplySearch()}}

\subsection*{Description}

Within the LALApplySearch Routine, the block search master receives the
first set of results from the slaves (processed within the LALApplySearch
on the slave node).  Then the master node either sends a data id of -1
signalling that all vectors that need processing have been sent out or the
master node send the index of the vector stored within the master's parameter
structure in a vector sequence "data".

On each slave node, as in the LALConditionData section, the slave first
processes the data vector that it received in LALConditionData.  Next the 
slave receives the id of the next vector and either exits if -1 is given or
waits to receive the next data vector if signalled by the master node.  At
the completetion of Rho2, each slave will exit.

The result values are stored in an array on the master node until all vectors
have been processed.  The results are reprocessed as dbEntries when received
from a slave node.  Upon receipt of all results, the master fills in the
"search summary" table and then attaches the chain of results.  This is done
using the \texttt{BlockBuildOutput} routine.

\subsection*{Algorithms}

The LALApplySearch routine does not manipulate the data in any way on either
master or the slave until the slave sends the data to the Rho2 routine.  The
LALApplySearch routine is designed as a wrapper to the facilitate 
parallelism mentioned earlier.  The Block Normal Search takes an input of a
vector sequence that can have any number of vectors in it.  It does make the
assumption that input vectors following a sampleRate scheme like that of wavelet
filtering - The last and second to last input segment should have identical
sample rates and every input previous to should have a sample rate higher by
a factor of two over the later segment.  For example:\\
\\
inout->sequence[0]->sampleRate = 4096\\
inout->sequence[1]->sampleRate = 2048\\
inout->sequence[2]->sampleRate = 1024\\
inout->sequence[3]->sampleRate = 512\\
inout->sequence[4]->sampleRate = 512\\
\\
This constitutes valid input.\\

The parallelism begins by breaking the input into individual vectors and sending each
vector to a different slave node.  Each slave node then preforms the Rho2
algorithm on its vector and returns the results

To note, the Rho2 algorithm is discussed below under the \texttt{BlockRho2}
routine heading.

The MPI exchange is as follows:
%
\begin{enumerate}
\item 
The Master waits to receive a message from a slave node that the slave
is ready to send its data.

\item 
The Slave then sends first the index of the data vector that it
processed.

\item 
Then the Slave sends its results

\item
The Master then sends back the index of the next vector to be processed
which is set to -1 if all data vectors have been processed.

\item
The Master then sends the next data vector to the slave, given that
there are still vectors to be processed.

\end{enumerate}

\subsection*{Uses}
 
\begin{itemize}
\item \texttt{LALMPISendMsg}
\item \texttt{LALMPIRecvMsg}
\item \texttt{LALMPISendINT4}
\item \texttt{LALMPIRecvINT4}
\item \texttt{LALMPISendREAL8}
\item \texttt{LALMPIRecvREAL8}
\item \texttt{LALDCreateVector}
\item \texttt{LALMPISendREAL8Vector}
\item \texttt{LALMPIRecvREAL8Vector}
\item \texttt{BlockRho2}
\item \texttt{BlockBuildOutput} 
\item \texttt{LALDDestroyVector}
\end{itemize}

\subsection*{Notes}

\vfill{\footnotesize\input{BlockApplySearchCV}}

 *************************************************** </lalLaTeX> */

#include "Block.h"

NRCSID( BLOCKAPPLYSEARCHC, "$Id: BlockApplySearch.c,v 1.30 2003/07/06 20:06:08 jolien Exp $");

/*  Function for local hidden function  */
int BlockCompare(const void* p1, const void* p2);


/************************ <lalVerbatim file="BlockApplySearchCP"> */
void LALApplySearch
(
	LALStatus		*status,
	LALSearchOutput		*output,
	LALSearchInput		*input,
	LALApplySearchParams	*applyParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declarations  */
	BlockSearchParams	*params;

	INITSTATUS( status, "BlockApplySearch", BLOCKAPPLYSEARCHC );
	ATTATCHSTATUSPTR (status);

	/*  Inititalize Variables  */
	params = ((BlockSearchParams*)(applyParams->searchParams));

	/*  Sanity Checks on Input Variables  */
	ASSERT( applyParams, status, BLOCKH_ENULL, BLOCKH_MSGENULL);

	/*  Get rank  */
	MPI_Comm_rank(*(applyParams->mpiParams->comm), &(params->rank));

	/*  Which node is this running on?  */
	if (params->searchMaster)
	{
		/* master */
		ApplySearchMaster(status->statusPtr, output, input, applyParams );
	}
	else
	{
		/* slave */
		ApplySearchSlave(status->statusPtr, output, input, applyParams );
	}

	DETATCHSTATUSPTR (status);
	RETURN(status);
}

/*
 * Function ApplySearchMaster
 *
 * Called from LALApplySearch only on the search master node. It's
 * main job is to wait for responses for each slave ie. retrieve the
 * statistics, send the next vector, wait for the slaves to end, and
 * then to format the output.
 */

/************************ <lalVerbatim file="BlockApplySearchMasterCP"> */
void ApplySearchMaster
(
	LALStatus		*status,
	LALSearchOutput		*output,
	LALSearchInput		*input,
	LALApplySearchParams	*applyParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declarations  */
	BlockSearchParams		*params;
	bandSegment			*tempPointer;
	bandObject			*bandSet;
	BlockResult			*current;
	REAL8Vector			*tempVector;
	MPIMessage			mpiMsg;
	MPI_Comm			*comm;
	UINT4				numberResults;
	UINT4				iterator;
	UINT4				level;

	/*  Initialize status to pass to other functions  */
	INITSTATUS( status, "DistributeDataMaster", BLOCKAPPLYSEARCHC );
	ATTATCHSTATUSPTR (status);

	/*  Sanity Checks on Variables  */
	ASSERT( input, status, BLOCKH_ENULL, BLOCKH_MSGENULL);
	ASSERT( output, status, BLOCKH_ENULL, BLOCKH_MSGENULL);

	/*  Variable Initialization  */
	comm	= ((MPI_Comm*)(applyParams->mpiParams->comm));
	params	= ((BlockSearchParams*)(applyParams->searchParams));
	current	= NULL;

	bandSet	= params->data;

	/*  Wait for Slaves to say, "I'm here"  */
	LALMPIRecvMsg( status->statusPtr, &mpiMsg, (*comm));
	CHECKSTATUSPTR( status );

	/*  Slave sends number of breakpoints */
	LALMPIRecvINT4(status->statusPtr, &(numberResults), mpiMsg.source, (*comm));
	CHECKSTATUSPTR( status );

#ifdef BLOCKSEARCH__DEBUG
	printf("\nNumber of Blocks found = %d\n",numberResults);
	params->eventCounter += numberResults;
#endif

	/*  This counts number of results still to get  */
	params->resultsObtained++;

	if( params->segmentsRemaining == 0 )
	{
		iterator = -1;

		/*  Send kill code  */
		LALMPISendINT4( status->statusPtr, &(iterator), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );
	}
	else	/*  Send next BandSet Object  */
	{
		REAL8  segmentLength;

		/*  Send bandSet->segmentID */
		iterator = (params->numberOfSequences - params->segmentsRemaining);
		LALMPISendINT4( status->statusPtr, &(iterator), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		segmentLength = ((REAL8)(bandSet->decomposition[0]->data->length)) / params->sampleRate[0];

		/*  Send decomposition[0]->data->length  */
		LALMPISendREAL8( status->statusPtr, &(segmentLength), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Send startTime  */
		LALMPISendINT4( status->statusPtr, &(bandSet->decomposition[0]->startTime), mpiMsg.source, (*comm));
		CHECKSTATUSPTR( status );

		/*  Send Data in for loop  */
		for(level = 0; level < bandSet->levels; level++ )
		{
			tempPointer	= NULL;
			tempVector	= NULL;

			tempPointer = bandSet->decomposition[level];

			/*  Allocate temporary vector  */
			LALDCreateVector(status->statusPtr, &tempVector, bandSet->decomposition[level]->data->length);
			CHECKSTATUSPTR( status );

			/*  Copy data to temp vector  */
			for( iterator = 0; iterator < tempVector->length; iterator++ )
			{
				tempVector->data[iterator] = bandSet->decomposition[level]->data->data[iterator];
			}

			/*  Send Data to Slave  */
			LALMPISendREAL8Vector(status->statusPtr, tempVector, mpiMsg.source, (*comm));
			CHECKSTATUSPTR( status );

			/*  Clean up temporary Mem each time used  */
			LALDDestroyVector(status->statusPtr, &tempVector);
			CHECKSTATUSPTR( status );

			/*  Remove bandSet Segment  */
			bandSet->decomposition[level] = ((bandSegment*)(bandSet->decomposition[level]->next));
			LALDDestroyVector(status->statusPtr, &tempPointer->data);
			CHECKSTATUSPTR( status );
			LALFree(tempPointer);

		}	/*  For loop to transmit data  */

		output->fracRemaining	= ( ((REAL8)(params->segmentsRemaining + params->numSlaves - 1)) / ((REAL8)(params->numberOfSequences)) );

		params->segmentsRemaining--;

	}	/*  ( params->segmentsRemaining != 0 )  */

	if( params->resultsObtained == params->numberOfSequences )
	{

#ifdef BLOCKSEARCH__DEBUG
	printf("\nTotal Number of Blocks Reported = %d\n",params->eventCounter);
#endif

	  BuildOutput( status->statusPtr, output, params );
	  output->fracRemaining	= 0;
	  output->notFinished	= 0;
	}

	/*  Normal Exit  */
	DETATCHSTATUSPTR( status );
	RETURN( status );
}

int BlockCompare(const void* p1, const void* p2)
{
    return ((*(*((BlockResult**)(p1)))).startTime - (*(*((BlockResult**)(p2)))).startTime);
} 


/*
 * Function ApplySearchSlave
 *
 * Called from LALApplySearch on each slave node. It receives the data
 * sent by the master node, processes it via the BlockRho2 function,
 * then it passes the results to the master node, and if there are more
 * vectors to be processed, it waits to receive the next one.
 */


/************************ <lalVerbatim file="BlockApplySearchMasterCP"> */
void ApplySearchSlave
(
	LALStatus				*status,
	LALSearchOutput			*output,
	LALSearchInput			*input,
	LALApplySearchParams	*applyParams
)
/*********************************************** </lalVerbatim> */
{
	/*  Variable Declarations  */
	BlockSearchParams	*params;
	BlockResult		*tempResult	= NULL;
	BlockResult		*tempResult2	= NULL;
	bandSegment		*tempPointer	= NULL;
	bandSegment		*current	= NULL;
	bandObject		*initData	= NULL;
	bandObject		*bandSet	= NULL;

	MPIMessage		sendMsg;
	MPI_Comm		*comm;

	REAL8			meanZeroSquared;
	REAL8			meanRecentered;
	REAL8			meanThreshold;
	REAL8			varThreshold;
	REAL8			segmentLength;

	INT4			startTime	= 0;
	INT4			dataStart	= 0;
	INT4			duration	= 0;
	INT4			iterator	= 0;
	INT4			length		= 0;
	INT4			level		= 0;
	INT4			index		= 0;

	/*  Initialize status to pass to other functions  */
	INITSTATUS( status, "DistributeDataSlave", BLOCKAPPLYSEARCHC );
	ATTATCHSTATUSPTR (status);

	/*  Variable Initialization  */
	comm	= ((MPI_Comm*)(applyParams->mpiParams->comm));
	params	= ((BlockSearchParams*)(applyParams->searchParams));
	bandSet	= params->data;

	/*  Create Copy of inital sequence to Reparse later  */
	initData = ((bandObject*)(LALMalloc( sizeof(bandObject))));
	if( !initData )
	  {
	    ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	  }

	/*  Must store copies of all levels of initial data  (One Data Vector per level)  */
	initData->decomposition = (bandSegment**)(LALMalloc ( sizeof(bandSegment*) * bandSet->levels ));
	if( !initData->decomposition)
	{
		ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	}

	/*  Block Normal Analysis of Data  */
	for( level = 0; level < params->data->levels; level++)
	{

	  /*  Need to store local copy of initial data to reparse for mean & variance  */
	  /*  According to breakpoints  */
	  initData->decomposition[level] = ((bandSegment*)(LALMalloc( sizeof(bandSegment))));
	  if( !initData->decomposition[level] )
	    {
	      ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	    }

	  initData->decomposition[level]->data = NULL;

     	  LALDCreateVector(status->statusPtr, &(initData->decomposition[level]->data), bandSet->decomposition[level]->data->length);
	  CHECKSTATUSPTR( status );

	  /*  Copy Initial Data Vector for Reparsing  */
	  for(length = 0; length < initData->decomposition[level]->data->length; length++)
	    {
	      initData->decomposition[level]->data->data[length] = bandSet->decomposition[level]->data->data[length];
	    }

	  /*  Store the index startTime to reference later  */
	  dataStart = bandSet->decomposition[level]->startTime;

	  BlockAddQ( status->statusPtr, params->data->decomposition[level], params );
	  params->data->decomposition[level] = NULL;
	}

	while(( current = BlockDelQ ( params ) ) != NULL )
	{
		/*  Send current data vector myParams->vec to BlockRho2 Function  */
		LALBlockRho2(status->statusPtr, &current->rhoN, &current->value, &current->index, current->data, &(params->marginOfExclusion));
		CHECKSTATUSPTR( status );

		/*  Function will split vector into two new segments to process if:  */
		/*  1. The Rho2 value is above threshold.  */
		/*  2. The resulting segment will have length greater than params->minVectorLength.  */
		BlockSplitLevel( status->statusPtr, current, params );
		CHECKSTATUSPTR( status );

		if ( current->left != NULL )
		{
			tempPointer = ((bandSegment*)(current->left));
			BlockAddQ ( status->statusPtr, tempPointer, params );
			CHECKSTATUSPTR( status );
		}

		if ( current->right != NULL )
		{
			tempPointer = ((bandSegment*)(current->right));
			BlockAddQ ( status->statusPtr, tempPointer, params );
			CHECKSTATUSPTR( status );
		}

		LALDDestroyVector( status->statusPtr, &current->data );
		LALFree( current );
	}

	/*  Format the result array into an output object to fill the database  */
	for( level = 0; level < bandSet->levels; level++ )
	  {
	    BlockResult	**temp;

	    /*  Allocate Array of pointers to sort  */
	    temp = (BlockResult**)(LALMalloc(sizeof(BlockResult*) * ( params->numberResults + 1)));
	    if( !temp )
	      {
		ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	      }

	    /*  Reset to first result  */
	    params->prevResults = params->results;

	    /*  Copy over pointers to each result in this level  */
	    for( iterator = 0; iterator < params->numberResults; iterator++ )
	      {
		temp[iterator] = params->prevResults;
		params->prevResults = ((BlockResult*)(params->prevResults->next));
	      }

	    if( params->numberResults > 0 )
	      {
		/*  Sort by startTime with the quicksort algorithm  */
		qsort(temp,params->numberResults,(sizeof(BlockResult*)),BlockCompare);
	      }

	    /*  Create blocks to output as results  */
	    tempPointer = ((bandSegment*)(LALMalloc(sizeof(bandSegment))));
	    if( !tempPointer )
	      {
		ABORT (status, BLOCKH_EALOC, BLOCKH_MSGEALOC);
	      }

	    /*  The beginning of the sequence is an implicit break point  */
	    /*  Initialize the beginning break point  */
	    tempPointer->duration = 0;
	    tempPointer->startTime = dataStart;
	    tempPointer->rhoN = -9999;

	    /*  Reparse to calculate individual block statistics  */
	    for( iterator = 0; iterator < params->numberResults; iterator++ )
	      {
		tempPointer->startTime += tempPointer->duration;

		tempPointer->duration = temp[iterator]->startTime - tempPointer->startTime;

		tempPointer->data = NULL;

		LALDCreateVector( status->statusPtr, &(tempPointer->data), tempPointer->duration);
		CHECKSTATUSPTR( status );

		/*  Copy Data for this block  */
		for( index = tempPointer->startTime - dataStart;  index < (tempPointer->startTime - dataStart + tempPointer->duration); index++ )
		  {
		    tempPointer->data->data[index - (tempPointer->startTime - dataStart)] = (REAL8)(initData->decomposition[level]->data->data[index]);
		  }

		/*  Calculate statistics for this block  */
		LALDMoment(status->statusPtr,&(temp[iterator]->variance), tempPointer->data, 2 );
		LALDMoment(status->statusPtr, &(temp[iterator]->mean), tempPointer->data, 1);

		temp[iterator]->startTime = tempPointer->startTime;
		temp[iterator]->duration = tempPointer->duration;

		/*  We must relink our list according to sorted order  */
		if( iterator < params->numberResults - 1 )
		  {
		    /*  Link to the next in order of start Times  */
		    ((BlockResult*)(temp[iterator]->next)) = temp[iterator + 1];
		  }
		else
		  {
		    /*  the end is also an implicit break point, so you must create a new block  */
		    ((BlockResult*)(temp[iterator]->next)) = tempResult = ((BlockResult*)(LALMalloc(sizeof(BlockResult))));
		  }

		temp[iterator]->rhoNEnd = temp[iterator]->rhoNStart;
		temp[iterator]->rhoNStart = tempPointer->rhoN;
		tempPointer->rhoN = temp[iterator]->rhoNEnd;

		temp[iterator]->segmentID = iterator;


#ifdef BLOCKSEARCH__DEBUG
		if(temp[iterator]->rhoNStart < 0)
		  fprintf(stdout,"bandSet(%d,%d)=\t%3.16f\t%3.16f\t%d\t%3.16f\t%3.16f\t%d\n", temp[iterator]->bandLevel, params->rank, temp[iterator]->rhoNStart, (REAL4)(temp[iterator]->rhoNEnd), temp[iterator]->bandLevel, temp[iterator]->mean, temp[iterator]->variance, temp[iterator]->startTime);
		else
		  fprintf(stdout,"bandSet(%d,%d)=\t%3.16f\t%3.16f\t%d\t%3.16f\t%3.16f\t%d\n", temp[iterator]->bandLevel, params->rank, (REAL4)(temp[iterator]->rhoNStart), (REAL4)(temp[iterator]->rhoNEnd), temp[iterator]->bandLevel, temp[iterator]->mean, temp[iterator]->variance, temp[iterator]->startTime);
		fflush(stdout);
#endif

		LALDDestroyVector( status->statusPtr, &(tempPointer->data));
		CHECKSTATUSPTR( status );
	      }

	    if( params->numberResults == 0 )
	      {
		temp[0] = tempResult = ((BlockResult*)(LALMalloc(sizeof(BlockResult))));
	      }

	    /*  Must fill in values for the final block because it had to be created from scratch.  */
	    tempResult->segmentID = iterator;
	    tempResult->index = 0;

	    tempPointer->startTime += tempPointer->duration;
	    tempPointer->duration = initData->decomposition[level]->data->length - (tempPointer->startTime - dataStart);
	    tempPointer->data = NULL;

	    LALDCreateVector( status->statusPtr, &(tempPointer->data), tempPointer->duration);
	    CHECKSTATUSPTR( status );

	    /*  Copy Data for this block  */
	    for( iterator = tempPointer->startTime - dataStart;  iterator < (tempPointer->startTime - dataStart + tempPointer->duration); iterator++ )
	      {
		tempPointer->data->data[iterator - (tempPointer->startTime - dataStart)] = (REAL8)(initData->decomposition[level]->data->data[iterator]);
	      }

	    /*  Calculate statistics for this block  */
	    LALDMoment(status->statusPtr,&(tempResult->variance), tempPointer->data, 2 );
	    LALDMoment(status->statusPtr, &(tempResult->mean), tempPointer->data, 1);

	    LALDDestroyVector( status->statusPtr, &(tempPointer->data));
	    CHECKSTATUSPTR( status );

	    tempResult->startTime = tempPointer->startTime;
	    tempResult->duration = tempPointer->duration;
	    tempResult->bandLevel = level;

	    tempResult->next = NULL;
	    tempResult->rhoNStart = tempPointer->rhoN;

	    /*  This is the dummy value for the final block's end  */
	    tempResult->rhoNEnd   = -9999;

	    /*  We added one block by adding the beginning and end both as implicit breakpoints  */
	    params->numberResults++;


#ifdef BLOCKSEARCH__DEBUG
	    if(tempResult->rhoNStart > 0)
	      fprintf(stdout,"bandSet(%d,%d)=\t%3.16f\t%3.16f\t%d\t%3.16f\t%3.16f\t%d\n", tempResult->bandLevel, params->rank, (REAL4)(tempResult->rhoNStart), tempResult->rhoNEnd, tempResult->bandLevel, tempResult->mean, tempResult->variance, (tempResult->startTime));
	    else
	      fprintf(stdout,"bandSet(%d,%d)=\t%3.16f\t%3.16f\t%d\t%3.16f\t%3.16f\t%d\n", tempResult->bandLevel, params->rank, tempResult->rhoNStart, tempResult->rhoNEnd, tempResult->bandLevel, tempResult->mean, tempResult->variance, (tempResult->startTime));
	    fflush(stdout);
#endif

	    /*  Mean & Variance Thresholding  */
	    params->results = NULL;
	    tempResult2 = NULL;

	    if( params->numberResults > 1 )
	      {
		tempResult = temp[0];
	      }

	    meanZeroSquared = pow(params->meanZero[level],2); 

	    /*  Set Mean Threshold */
	    meanThreshold   = params->meanThreshold[level] * params->varianceZero[level] + params->meanOffset[level] * meanZeroSquared;

	    /*  Set Variance Threshold */
	    varThreshold    = params->varianceThreshold[level] * params->varianceZero[level];

	    while( tempResult->next != NULL )
	      {
		meanRecentered  = (tempResult->mean - params->meanZero[level]);
		meanRecentered  = pow(meanRecentered,2);

		/* Is mean greater than threshold? */
		if ( meanRecentered >= meanThreshold )
		  {
		    if( tempResult2 == NULL )
		      {
			params->results = tempResult;
			tempResult2 = tempResult;
		      }
		    else
		      {
			((BlockResult*)(tempResult2->next)) = tempResult;
			tempResult2 = ((BlockResult*)(tempResult2->next));
		      }
		    tempResult = ((BlockResult*)(tempResult->next));
		  }
		else
		  {
		  /*  Is variance greater than threshold? */
		    if ( tempResult->variance > varThreshold)
		      {
			if( tempResult2 == NULL )
			  {
			    params->results = tempResult;
			    tempResult2 = tempResult;
			  }
			else
			  {
			    ((BlockResult*)(tempResult2->next)) = tempResult;
			    tempResult2 = ((BlockResult*)(tempResult2->next));
			  }
			
			tempResult = ((BlockResult*)(tempResult->next));
		      }
		    else
		      {
			/*  Result is not greater than either threshold, so delete it. */
			BlockResult*    del;

			del = tempResult;

			tempResult = ((BlockResult*)(tempResult->next));
			LALFree(del);

			params->numberResults--;
		      }
		  }
	      }

	    meanRecentered  = (tempResult->mean - params->meanZero[level]);
	    meanRecentered  = pow(meanRecentered,2);

	    /* Is mean greater than threshold? */
	    if ( meanRecentered >= meanThreshold )
	      {
		if( tempResult2 == NULL )
		  {
		    params->results = tempResult;
		    tempResult2 = tempResult;
		  }
		else
		  {
		    ((BlockResult*)(tempResult2->next)) = tempResult;
		    tempResult2 = ((BlockResult*)(tempResult2->next));
		  }
	      }
	    else
	      {
		/*  Is variance greater than threshold? */
		if ( tempResult->variance > varThreshold)
		  {
		    if( tempResult2 == NULL )
		      {
			params->results = tempResult;
			tempResult2 = tempResult;
		      }
		    else
		      {
			((BlockResult*)(tempResult2->next)) = tempResult;
			tempResult2 = ((BlockResult*)(tempResult2->next));
		      }
		  }
		else
		  {
		    /*  Result is not greater than either threshold, so delete it. */
		    BlockResult*    del;

		    del = tempResult;
		    LALFree(del);

		    params->numberResults--;
		  }
	      }
 
	    if( tempResult2 != NULL )
	      {
		tempResult2->next = NULL;
	      }

	    if( tempPointer != NULL)
	      {
		LALFree(tempPointer);
	      }

	    if( temp != NULL )
	      {
		LALFree(temp);
	      }
	  }

	/*  Set up MPI Message  */
	sendMsg.source	= params->rank;
	sendMsg.msg	= PassData;
	sendMsg.send	= 1;

	/*  Send Message to say I am ready  */
	LALMPISendMsg( status->statusPtr, &sendMsg, 0, (*comm));
	CHECKSTATUSPTR( status );

	/*  Slave sends number of results */
	LALMPISendINT4(status->statusPtr, &(params->numberResults), 0, (*comm));
	CHECKSTATUSPTR( status );

	if( params->numberResults )
	  {
	    /* Chain results just received onto stored results  */
	    AddResults(status->statusPtr, output, params);
	  }

	params->head = params->tail = NULL;
	current = NULL;

	/*  Get rid of local copy of data  */
	for( level = 0;  level < bandSet->levels; level++)
	  {
	    LALDDestroyVector(status->statusPtr, &(initData->decomposition[level]->data));
	    LALFree(initData->decomposition[level]);
	  }

	LALFree( initData->decomposition );
	LALFree( initData );
	
	/*  Here we have completely processed a sequence and returned the results  */
	/*  Now we send a message to ask if there is more data to work on  */

	/*  Receive bandSet->segmentID */
	LALMPIRecvINT4( status->statusPtr, &(bandSet->segmentID), 0, (*comm));
	CHECKSTATUSPTR( status );

	params->eventCounter = params->maxVectorLength * params->sampleRate[0] * bandSet->segmentID * bandSet->levels;

	/*  Are we done??  */
	if( bandSet->segmentID < 0 )
	{
	  /*  This halts ApplySearch on the slave node  */
	  output->fracRemaining	= 0;
	  output->notFinished 	= 0;

	  DETATCHSTATUSPTR( status );
	  RETURN( status );
	}

	/*  Reset this for next segment  */
	params->numberResults = 0;

	/*  Receive decomposition[0]->data->length  */
	LALMPIRecvREAL8( status->statusPtr, &(segmentLength), 0, (*comm));
	CHECKSTATUSPTR( status );

	/*  Receive startTime  */
	LALMPIRecvINT4( status->statusPtr, &(startTime), 0, (*comm));
	CHECKSTATUSPTR( status );

	for( level = 0; level < bandSet->levels; level++)
	{
		params->results = NULL;
		params->prevResults = NULL;

		bandSet->decomposition[level] = current = (bandSegment*)( LALMalloc( sizeof(bandSegment)));

		current->rhoN = current->index = current->value = 0;
		current->left	= NULL;
		current->right	= NULL;
		current->data	= NULL;
		current->next	= NULL;

		current->startTime = startTime;
		current->duration = duration;

		current->level = level;

		LALDCreateVector( status->statusPtr, &(current->data), segmentLength * params->sampleRate[level]);

		/*  Receive data from Master  */
		LALMPIRecvREAL8Vector(status->statusPtr, current->data, 0, (*comm));
		CHECKSTATUSPTR( status );
	}

	/*  Normal Exit  */
	DETATCHSTATUSPTR( status );
	RETURN( status );
}
