/********************************** <lalVerbatim file="ExtTrigDataCV">
Author: Rahkola, R.
$Id: ExtTrigData.c,v 1.8 2003/07/03 18:30:56 jolien Exp $
**************************************************** </lalVerbatim> */

/******************************* <lalLaTeX file="ExtTrigDataC">
\subsection{Module \texttt{ExtTrigData.c}}
\label{s:ExtTrigData.c}

Routines for creating and destroying a 
\texttt{ETDataSegmentVector}.

\subsubsection*{Prototypes}
\input{ExtTrigDataCP}
\index{\texttt{LALCreateETDataSegmentVector()}}

\subsection*{Description}

Creates and destroys a \texttt{ETDataSegmentVector},
which holds the input time-series data for two interferometers and a data
quality channel.

\subsection*{Algorithms}
  
\subsection*{Uses}

\begin{itemize}
\item \texttt{LALMalloc}
\item \texttt{LALCalloc}
\item \texttt{LALFree}
\item \texttt{LALCreateVector}
\item \texttt{LALDestroyVector}
\item \texttt{LALCCreateVector}
\item \texttt{LALCDestroyVector}
\end{itemize}

\subsection*{Notes}

\vfill{\footnotesize\input{ExtTrigDataCV}}

*************************************************** </lalLaTeX> */

#include <lal/LALStdlib.h>
#include <lal/SeqFactories.h>
#include "ExtTrig.h"

NRCSID (EXTTRIGDATAC, "$Id: ExtTrigData.c,v 1.8 2003/07/03 18:30:56 jolien Exp $");

/************************ <lalVerbatim file="ExtTrigDataCP"> */
void LALCreateETDataSegmentVector (
				   LALStatus            *status,
				   ETDataSegmentVector **vector,
				   ETInitParams         *params
)
/*********************************************** </lalVerbatim> */
{
    
  INT4                    i;              /* iterator for 'for' loops        */
  UINT4                   numPoints;      /* number of points in timeSeries  */
  UINT4                   filtLength;     /* number of points in freqSeries  */
  ETDataSegmentVector    *vectorPtr;      /* pointer to 'vector' arg         */
  ETDataSegment          *segPtr;         /* pointer to index in 'vector'    */
    
  INITSTATUS (status, "LALCreateETDataSegmentVector", EXTTRIGDATAC);
  ATTATCHSTATUSPTR (status);

  /* 
   *
   * check output vector and parameters 
   *
   */
  
  ASSERT ( !*vector, status, EXTTRIGDATAH_ENNUL, EXTTRIGDATAH_MSGENNUL );
  ASSERT (   params, status, EXTTRIGDATAH_ENULL, EXTTRIGDATAH_MSGENULL );
/*  if ( params->numSegments <= 0 ) {
      ABORT ( status, EXTTRIGDATAH_ENSEG, EXTTRIGDATAH_MSGENSEG );
    }
  if (params->numPoints <= 0) { 
      ABORT ( status, EXTTRIGDATAH_ENPTS, EXTTRIGDATAH_MSGENPTS );
    }
  
  if (params->numFrequencies <= 0) {
      ABORT (status, 
	     EXTTRIGDATAH_ENFRQ, 
	     EXTTRIGDATAH_MSGENFRQ);
    }
*/  
  /* vector length */
  numPoints = params->segLength;
  filtLength = params->segLength / 2 + 1;

  /**********************
   **********************
   *** allocate memory: note that TSeries and FSeries are placeholders for
   *** params->tsChan1, params->tsChan2, and params->fsKernel 
   **********************
   **********************/
  /* pointer to data vector */
  vectorPtr = *vector = (ETDataSegmentVector *) 
    LALMalloc (sizeof (ETDataSegmentVector));
  if ( !vector ) {
      ABORT (status, EXTTRIGDATAH_EALOC, EXTTRIGDATAH_MSGEALOC );
    }
  
  /* assign vector length, segsToBeAnalyzed, and segsAnalyzed */ 
  vectorPtr->length = params->numOffSegments + 1;
  vectorPtr->segsToBeAnalyzed = vectorPtr->length;
  vectorPtr->segsAnalyzed     = 0;
#if DEBUG_EXTTRIG > 2
  fprintf( stdout, "Creating ETDataSegmentVector,\tlength %d \n",
	   vectorPtr->length );
  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 2*/

  /* pointer to data */
  segPtr = vectorPtr->data = (ETDataSegment *) 
    LALCalloc (vectorPtr->length, sizeof (ETDataSegment));
  if (!segPtr)
    {
      /* is LALFree() really needed here?? */
/*        LALFree( vectorPtr ); */
/*        vectorPtr = NULL; */
      ABORT( status, EXTTRIGDATAH_EALOC, EXTTRIGDATAH_MSGEALOC );
    }
  
  /* individual data segments */ 
  for (i = 0; i < vectorPtr->length ; ++i)
    {
      /* tsChan1: don't need to allocate REAL4 array; just points to the
       * correct spot in the base REAL4TimeSeries
       */
      segPtr[i].tsChan1 = (REAL4TimeSeries *) 
	LALCalloc ( 1, sizeof (REAL4TimeSeries) );
      if (!(segPtr[i].tsChan1))
        {
	  ABORT (status, EXTTRIGDATAH_EALOC, EXTTRIGDATAH_MSGEALOC);
        }
      segPtr[i].tsChan1->data = (REAL4Sequence *)
	LALMalloc( sizeof(REAL4Sequence) );
      if (!(segPtr[i].tsChan1->data)) {
	ABORT (status, EXTTRIGDATAH_EALOC, EXTTRIGDATAH_MSGEALOC);
      }
      segPtr[i].tsChan1->data->length = numPoints;
      segPtr[i].tsChan1->data->data = NULL;
      
      /* tsChan2: don't need to allocate REAL4 array; just points to the
       * correct spot in the base REAL4TimeSeries
       */
      segPtr[i].tsChan2 = (REAL4TimeSeries *) 
	LALCalloc ( 1, sizeof (REAL4TimeSeries) );
      if (!(segPtr[i].tsChan2))
        {
	  ABORT (status, EXTTRIGDATAH_EALOC, EXTTRIGDATAH_MSGEALOC);
        }
      segPtr[i].tsChan2->data = (REAL4Sequence *)
	LALMalloc( sizeof(REAL4Sequence) );
      if (!(segPtr[i].tsChan2->data)) {
	ABORT (status, EXTTRIGDATAH_EALOC, EXTTRIGDATAH_MSGEALOC);
      }
      segPtr[i].tsChan2->data->length = numPoints;
      segPtr[i].tsChan2->data->data = NULL;

      /* filter kernel is no longer needed on a per-segment basis;
       * merely referencing the main filter kernel is sufficient
       */
      segPtr[i].filter = NULL;

      /* notch is no longer needed on a per-segment basis;
       * merely referencing the main notch is sufficient
       */
      segPtr[i].notch = NULL;

      /* fftPlan is no longer needed on a per-segment basis;
       * merely referencing the main fftPlan is sufficient
       */
      segPtr[i].fftPlan = NULL;
    }
  
  /* normal exit */
  DETATCHSTATUSPTR (status);
  RETURN (status);
}   


/************************ <lalVerbatim file="ExtTrigDataCP"> */
void LALDestroyETDataSegmentVector(
				   LALStatus*              status,
				   ETDataSegmentVector**   vector
)
/*********************************************** </lalVerbatim> */
{
  INT4                    i;
  ETDataSegment          *segPtr;
  
  
  INITSTATUS (status, "LALDestroyETDataSegmentVector", EXTTRIGDATAC);
  ATTATCHSTATUSPTR (status);
  
  
  /* 
   *
   * check that there is something that needs to be destroyed 
   *
   */
  ASSERT (*vector, status, EXTTRIGDATAH_ENULL, EXTTRIGDATAH_MSGENULL);
  segPtr = (*vector)->data;
  
  
  /*
   *
   * free memory
   *
   */
#if DEBUG_EXTTRIG > 2
  fprintf( stdout, "Destroying ETDataSegmentVector \n" );
  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 2*/
  /* should all pointers be set to NULL after LALFree?? */
  for (i = 0; i < (*vector)->length; ++i)
    {
      /* don't have to free filter, as it's just a pointer to fsKernel */
      /* note that freeing of REAL4Vectors will be done in FinalizeSearch */
      LALFree( segPtr[i].tsChan2->data );
      LALFree( segPtr[i].tsChan2 );

      LALFree( segPtr[i].tsChan1->data );
      LALFree( segPtr[i].tsChan1 );
    }
  
  LALFree( segPtr );
  LALFree( *vector );
  *vector = NULL;
  
  /* normal exit */
  DETATCHSTATUSPTR (status);
  RETURN (status);
}


/************************ <lalVerbatim file="ExtTrigDataCP"> */
void LALCreateETSegList (
			 LALStatus     *status,
			 ETSegList    **list,
			 ETInitParams  *params )
/*********************************************** </lalVerbatim> */
{
  UINT4                   window;     /* length of the window (in counts)*/
  UINT4                   listLength; /* number of segments in list      */
  ETSegList              *listPtr;    /* pointer to 'list' arg           */
  ETSegment              *segPtr;     /* pointer to index in 'list'      */
    
  INITSTATUS (status, "LALCreateETSegList", EXTTRIGDATAC);
  ATTATCHSTATUSPTR (status);

  /* 
   * check output vector and parameters 
   */
  
  ASSERT ( !*list, status, EXTTRIGDATAH_ENNUL, EXTTRIGDATAH_MSGENNUL );
  ASSERT (   params, status, EXTTRIGDATAH_ENULL, EXTTRIGDATAH_MSGENULL );
  if ( params->numOffSegments <= 0 ) {       /* negative number of segments */
    ABORT ( status, EXTTRIGDATAH_ENSEG, EXTTRIGDATAH_MSGENSEG );
  }
  if ( params->numOffSegments >
       ((params->offTimeShiftHigh - params->offTimeShiftLow) / 
	(params->segLength + 2 * params->segMaxDelay)) ){ /*too many segments*/
    ABORT( status, EXTTRIGDATAH_ETOTL, EXTTRIGDATAH_MSGETOTL );
  }

  /* vector length = (time available for segments)*/
  window = (params->segLength);
  listLength = ( (params->offTimeShiftHigh - params->offTimeShiftLow) / 
		 (params->segLength + 2 * params->segMaxDelay) ) + 1;
#if DEBUG_EXTTRIG > 2
  fprintf( stdout, "Creating ETSegList,\tlength %d \n", listLength );
#endif /*  DEBUG_EXTTRIG > 2*/
#if DEBUG_EXTTRIG > 1
  fprintf( stdout, 
	   "Number of possible segments: [(%d - %d) / (%d + %d)] + 1 = %d \n",
	   params->offTimeShiftHigh, params->offTimeShiftLow, 
	   window, 2 * params->segMaxDelay, listLength       );
  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 1*/

  /*
   * allocate memory 
   */

  /* pointer to segment list */
  listPtr = *list = (ETSegList *) LALMalloc( sizeof(ETSegList) );
  if ( !list ) {
      ABORT (status, EXTTRIGDATAH_EALOC, EXTTRIGDATAH_MSGEALOC );
    }
  
  /* assign list length and filled-segments */ 
  listPtr->length = listLength;
  listPtr->filled = 0;
  
  /* pointer to segments */
  segPtr = listPtr->data = (ETSegment *) 
    LALCalloc( listPtr->length, sizeof (ETSegment) );
  if (!segPtr)
    {
      /* is LALFree() really needed here?? */
/*        LALFree( listPtr ); */
/*        listPtr = NULL; */
      ABORT( status, EXTTRIGDATAH_EALOC, EXTTRIGDATAH_MSGEALOC );
    }
  
  /* normal exit */
  DETATCHSTATUSPTR (status);
  RETURN (status);
}   


/************************ <lalVerbatim file="ExtTrigDataCP"> */
void LALDestroyETSegList(
			 LALStatus  *status,
			 ETSegList **list   )
/*********************************************** </lalVerbatim> */
{
  ETSegment    *segPtr;

  INITSTATUS( status, "LALDestroyETSegList", EXTTRIGDATAC );
  ATTATCHSTATUSPTR( status );
  
  /* 
   * check that there is something that needs to be destroyed 
   */
  ASSERT( *list, status, EXTTRIGDATAH_ENULL, EXTTRIGDATAH_MSGENULL );
  segPtr = (*list)->data;

#if DEBUG_EXTTRIG > 2
  fprintf( stdout, "Destroying ETSegList \n" );
  fflush( stdout );
#endif /*  DEBUG_EXTTRIG > 2*/

  LALFree( segPtr );
  LALFree( *list );
  *list = NULL;
  
  /* normal exit */
  DETATCHSTATUSPTR( status );
  RETURN( status );
}
