/********************************** <lalVerbatim file="StochasticDataCV">
Author: Anderson, W.G., Romano, J.D., and Whelan, J.T.
$Id: StochasticData.c,v 1.8 2002/08/13 21:16:26 whelan Exp $
**************************************************** </lalVerbatim> */

/******************************* <lalLaTeX file="StochasticDataC">
\subsection{Module \texttt{StochasticData.c}}
\label{s:StochasticData.c}

Routines for creating and destroying a
\texttt{StochasticRealDataSegmentVector} or
\texttt{StochasticComplexDataSegmentVector}

\subsubsection*{Prototypes}
\input{StochasticDataCP}
\index{\texttt{LALSCreateStochasticDataSegmentVector()}}
\index{\texttt{LALCCreateStochasticDataSegmentVector()}}

\subsection*{Description}

Creates and destroys a \texttt{StochasticRealDataSegmentVector} or
\texttt{StochasticComplexDataSegmentVector}, which holds the input
time-series data for two interferometers.

\subsection*{Algorithms}
  
\subsection*{Uses}

\begin{itemize}
\item \texttt{LALMalloc}
\item \texttt{LALFree}
\item \texttt{LALSCreateVector}
\item \texttt{LALSDestroyVector}
\item \texttt{LALCCreateVector}
\item \texttt{LALCDestroyVector}
\end{itemize}

\subsection*{Notes}

\vfill{\footnotesize\input{StochasticDataCV}}

*************************************************** </lalLaTeX> */

#include <lal/LALStdlib.h>
#include <lal/SeqFactories.h>
#include "StochasticData.h"

NRCSID (STOCHASTICDATAC, "$Id: StochasticData.c,v 1.8 2002/08/13 21:16:26 whelan Exp $");

/************************ <lalVerbatim file="StochasticDataCP"> */
void LALSCreateStochasticDataSegmentVector (
    LALStatus*                      status,
    StochasticRealDataSegmentVector**   vector,
    StochasticInitParams*           params
)
/*********************************************** </lalVerbatim> */
{
    
    INT4                            i;
    StochasticRealDataSegmentVector*    vectorPtr;
    StochasticRealDataSegment*          segPtr;

    
    INITSTATUS (status, 
                "LALSCreateStochasticDataSegmentVector", 
                STOCHASTICDATAC);
    
    ATTATCHSTATUSPTR (status);


    /* 
     *
     * check output vector and parameters 
     *
     */
    
    ASSERT (!*vector, 
            status, 
            STOCHASTICDATAH_ENNUL, 
            STOCHASTICDATAH_MSGENNUL);

    ASSERT (params, 
            status, 
            STOCHASTICDATAH_ENULL, 
            STOCHASTICDATAH_MSGENULL);
    
    if (params->numSegments <= 0)
    {
        ABORT (status,
               STOCHASTICDATAH_ENSEG, 
               STOCHASTICDATAH_MSGENSEG);
    }
    
    if (params->numPoints <= 0)
    { 
        ABORT (status, 
               STOCHASTICDATAH_ENPTS, 
               STOCHASTICDATAH_MSGENPTS);
    }

    if (params->numFrequencies <= 0)
    {
        ABORT (status, 
               STOCHASTICDATAH_ENFRQ, 
               STOCHASTICDATAH_MSGENFRQ);
    }


    /*
     *
     * allocate memory 
     *
     */
    
    /* pointer to vector */

    vectorPtr = *vector = (StochasticRealDataSegmentVector *) 
        LALMalloc (sizeof (StochasticRealDataSegmentVector));
    if (!vector)
    {
        ABORT (status, 
               STOCHASTICDATAH_EALOC, 
               STOCHASTICDATAH_MSGEALOC );
    }
   
    /* assign vector length */ 
    vectorPtr->length = params->numSegments;

    
    /* pointer to segment */

    segPtr = vectorPtr->data = (StochasticRealDataSegment *) 
        LALMalloc (params->numSegments * sizeof (StochasticRealDataSegment));
    if (!segPtr)
    {
        /* :TODO: is LALFree() really needed here?? */
        LALFree (vectorPtr);
        vectorPtr = NULL;
        ABORT (status, 
               STOCHASTICDATAH_EALOC, 
               STOCHASTICDATAH_MSGEALOC);
    }

    /* individual data segments */ 

    for (i = 0; i < vectorPtr->length ; ++i)
    {
        /* data1 */
        segPtr[i].data1 = (REAL4TimeSeries *) 
            LALMalloc (sizeof (REAL4TimeSeries));
        if (!(segPtr[i].data1))
        {
            ABORT (status,
                   STOCHASTICDATAH_EALOC, 
                   STOCHASTICDATAH_MSGEALOC);
        }
        
        segPtr[i].data1->data = NULL;
        TRY( LALSCreateVector (status->statusPtr, 
			       &(segPtr[i].data1->data), 
			       params->numPoints),
	     status );

        /* data2 */
        segPtr[i].data2 = (REAL4TimeSeries *) 
            LALMalloc (sizeof (REAL4TimeSeries));
        if (!(segPtr[i].data2))
        {
            ABORT (status,
                   STOCHASTICDATAH_EALOC, 
                   STOCHASTICDATAH_MSGEALOC);
        }
        
        segPtr[i].data2->data = NULL;
        TRY( LALSCreateVector (status->statusPtr, 
			       &(segPtr[i].data2->data), 
			       params->numPoints),
	     status );
    }

    /* normal exit */
    DETATCHSTATUSPTR (status);
    RETURN (status);
}   


/************************ <lalVerbatim file="StochasticDataCP"> */
void LALSDestroyStochasticDataSegmentVector(
    LALStatus*                      status,
    StochasticRealDataSegmentVector**   vector
)
/*********************************************** </lalVerbatim> */
{
    
    INT4                    i;
    StochasticRealDataSegment*  segPtr;

    
    INITSTATUS (status, 
                "LALSDestroyStochasticDataSegmentVector", 
                STOCHASTICDATAC);

    ATTATCHSTATUSPTR (status);

    
    /* 
     *
     * check that there is something that needs to be destroyed 
     *
     */
    
    ASSERT (*vector, 
            status, 
            STOCHASTICDATAH_ENULL, 
            STOCHASTICDATAH_MSGENULL);

    segPtr = (*vector)->data;

    
    /*
     *
     * free memory
     *
     */

    /* :TODO: should all pointers be set to NULL after LALFree?? */
    
    for (i = 0; i < (*vector)->length; ++i)
    {
      TRY( LALSDestroyVector (status->statusPtr, 
			      &(segPtr[i].data2->data)),
	   status );
      LALFree (segPtr[i].data2);
      
      TRY( LALSDestroyVector (status->statusPtr, 
			      &(segPtr[i].data1->data)),
	   status );
      LALFree (segPtr[i].data1);
    }

    LALFree (segPtr);

    LALFree (*vector);
    *vector = NULL;

    /* normal exit */
    DETATCHSTATUSPTR (status);
    RETURN (status);
}


/************************ <lalVerbatim file="StochasticDataCP"> */
void LALCCreateStochasticDataSegmentVector (
    LALStatus*                      status,
    StochasticComplexDataSegmentVector**   vector,
    StochasticInitParams*           params
)
/*********************************************** </lalVerbatim> */
{
    
    INT4                            i;
    StochasticComplexDataSegmentVector*    vectorPtr;
    StochasticComplexDataSegment*          segPtr;

    
    INITSTATUS (status, 
                "LALCCreateStochasticDataSegmentVector", 
                STOCHASTICDATAC);
    
    ATTATCHSTATUSPTR (status);


    /* 
     *
     * check output vector and parameters 
     *
     */
    
    ASSERT (!*vector, 
            status, 
            STOCHASTICDATAH_ENNUL, 
            STOCHASTICDATAH_MSGENNUL);

    ASSERT (params, 
            status, 
            STOCHASTICDATAH_ENULL, 
            STOCHASTICDATAH_MSGENULL);
    
    if (params->numSegments <= 0)
    {
        ABORT (status,
               STOCHASTICDATAH_ENSEG, 
               STOCHASTICDATAH_MSGENSEG);
    }
    
    if (params->numPoints <= 0)
    { 
        ABORT (status, 
               STOCHASTICDATAH_ENPTS, 
               STOCHASTICDATAH_MSGENPTS);
    }

    if (params->numFrequencies <= 0)
    {
        ABORT (status, 
               STOCHASTICDATAH_ENFRQ, 
               STOCHASTICDATAH_MSGENFRQ);
    }


    /*
     *
     * allocate memory 
     *
     */
    
    /* pointer to vector */

    vectorPtr = *vector = (StochasticComplexDataSegmentVector *) 
        LALMalloc (sizeof (StochasticComplexDataSegmentVector));
    if (!vector)
    {
        ABORT (status, 
               STOCHASTICDATAH_EALOC, 
               STOCHASTICDATAH_MSGEALOC );
    }
   
    /* assign vector length */ 
    vectorPtr->length = params->numSegments;

    
    /* pointer to segment */

    segPtr = vectorPtr->data = (StochasticComplexDataSegment *) 
        LALMalloc (params->numSegments * sizeof (StochasticComplexDataSegment));
    if (!segPtr)
    {
        /* :TODO: is LALFree() really needed here?? */
        LALFree (vectorPtr);
        vectorPtr = NULL;
        ABORT (status, 
               STOCHASTICDATAH_EALOC, 
               STOCHASTICDATAH_MSGEALOC);
    }

    /* individual data segments */ 

    for (i = 0; i < vectorPtr->length ; ++i)
    {
        /* data1 */
        segPtr[i].data1 = (COMPLEX8TimeSeries *) 
            LALMalloc (sizeof (COMPLEX8TimeSeries));
        if (!(segPtr[i].data1))
        {
            ABORT (status,
                   STOCHASTICDATAH_EALOC, 
                   STOCHASTICDATAH_MSGEALOC);
        }
        
        segPtr[i].data1->data = NULL;
        TRY( LALCCreateVector (status->statusPtr, 
			       &(segPtr[i].data1->data), 
			       params->numPoints),
	     status );

        /* data2 */
        segPtr[i].data2 = (COMPLEX8TimeSeries *) 
            LALMalloc (sizeof (COMPLEX8TimeSeries));
        if (!(segPtr[i].data2))
        {
            ABORT (status,
                   STOCHASTICDATAH_EALOC, 
                   STOCHASTICDATAH_MSGEALOC);
        }
        
        segPtr[i].data2->data = NULL;
        TRY( LALCCreateVector (status->statusPtr, 
			       &(segPtr[i].data2->data), 
			       params->numPoints),
	     status );
    }

    /* normal exit */
    DETATCHSTATUSPTR (status);
    RETURN (status);
}   


/************************ <lalVerbatim file="StochasticDataCP"> */
void LALCDestroyStochasticDataSegmentVector(
    LALStatus*                      status,
    StochasticComplexDataSegmentVector**   vector
)
/*********************************************** </lalVerbatim> */
{
    
    INT4                    i;
    StochasticComplexDataSegment*  segPtr;

    
    INITSTATUS (status, 
                "LALCDestroyStochasticDataSegmentVector", 
                STOCHASTICDATAC);

    ATTATCHSTATUSPTR (status);

    
    /* 
     *
     * check that there is something that needs to be destroyed 
     *
     */
    
    ASSERT (*vector, 
            status, 
            STOCHASTICDATAH_ENULL, 
            STOCHASTICDATAH_MSGENULL);

    segPtr = (*vector)->data;

    
    /*
     *
     * free memory
     *
     */

    /* :TODO: should all pointers be set to NULL after LALFree?? */
    
    for (i = 0; i < (*vector)->length; ++i)
    {
      TRY( LALCDestroyVector (status->statusPtr, 
			      &(segPtr[i].data2->data)),
	   status );
      LALFree (segPtr[i].data2);

      TRY( LALCDestroyVector (status->statusPtr, 
			      &(segPtr[i].data1->data)),
	   status );
      LALFree (segPtr[i].data1);
    }

    LALFree (segPtr);

    LALFree (*vector);
    *vector = NULL;

    /* normal exit */
    DETATCHSTATUSPTR (status);
    RETURN (status);
}

